; ProtMode.ASM

;--------------
; Useful externals

              EXTRN     C pGDT:WORD

;--------------
; Set up assembly conditions and pull in useful definitions

              IDEAL                         ; use Borland Ideal syntax
              P386                          ; enable protected mode code
              LOCALS

              MODEL     SMALL,C             ; C language calling sequence

              INCLUDE   "..\ffts.inc"
              INCLUDE   "..\firmdev.inc"

;------------------------
; Make the transition to protected mode

              CODESEG

              PROC      PMEntry
              ARG       VectorA:WORD,VectorB:WORD
              PUBLIC    PMEntry

              MOV       AL,001h             ; flag entry here
              MOV       DX,SYNC_ADDR
              OUT       DX,AL

              MOV       AX,NOT 0101h        ; preset the digits to dashes
              MOV       DX,LED_ADDR
              OUT       DX,AX

;--- wait for diskette motor to stop spinning

              MOV       AX,0040h            ; aim at motor status
              MOV       ES,AX
              MOV       SI,003Fh            ;  at 0040:003F

@@Motors:
              MOV       AL,[ES:SI]
              TEST      AL,03h              ; drives A & B
              JNZ       @@Motors

;--- set up the registers

              MOV       AX,DS               ; set ES:SI to GDT base address
              MOV       ES,AX
              MOV       SI,[pGDT]
              MOV       BH,[BYTE PTR VectorA]         ; IRQ 0:7
              MOV       BL,[BYTE PTR VectorB]         ; IRQ 8:15

              MOV       AL,003h             ; mark start of BIOS function
              OUT       DX,AL

;--- real mode

              MOV       AH,89h              ; BIOS function
              INT       15h                 ; make the transition

;--- protected mode?

              MOV       AL,080h             ; mark just after the function
              MOV       DX,SYNC_ADDR
              OUT       DX,AL

;--- decide what mode we're in

              JNC       @@PM

;--- still in real mode, so flag the failure

              XOR       CX,CX
              MOV       DX,SYNC_ADDR
@@Fail:
              IN        AL,DX               ; by blinking the whole port
              NOT       AL
              OUT       DX,AL

              MOV       BL,5
@@Fail1:      LOOP      @@Fail1             ; delay for a while
              DEC       BL
              JNZ       @@Fail1

              JMP       @@Fail

;--- protected mode!

@@PM:
              XOR       CX,CX               ; initialize the delay counter

              MOV       DX,LED_ADDR         ; show a couple of Ps
              MOV       AX,NOT 6767h
              OUT       DX,AX

@@Pass:
              MOV       DX,SYNC_ADDR        ; blink bit 6 on the parallel port
              IN        AL,DX
              XOR       AL,40h
              OUT       DX,AL

              MOV       BL,5
@@Punt:       LOOP      @@Punt              ; delay for a while
              DEC       BL
              JNZ       @@Punt

              MOV       DX,SW_ADDR          ; check switches to blow up
              IN        AX,DX
              NOT       AX                  ;  ... ON = 1
              MOV       ES,AX               ; GP fault = 13 decimal = 0D hex
                                            ;  if AX is not valid selector

              JMP       @@Pass

              ENDP      PMEntry


;------------------------
; Bad interrupt revectoring entry points
; Each IDT entry selects one of these vectors so we know how we got here
; They're padded to eight bytes to make life easier when we hit 32-bit mode

              PROC      PMBadIntVector
              PUBLIC    PMBadIntVector

@@ID          =         0
              REPT      256
              PUSH      LOW @@ID
              JMP       PMBadIntHandler
              NOP
              NOP
              NOP
@@ID          =         @@ID + 1
              ENDM

              ENDP      PMBadIntVector

;------------------------
; Bad interrupt handler
; Some IDT entries must be task gates, but we ignore that for now...
;  all IDT entries are interrupt gates, so interrupts are disabled here
; The response here is to show the interrupt ID and blink the decimals
;  we come from the revectoring jumps, so the DWORD on the stack is the ID

              PROC      PMBadIntHandler
              PUBLIC    PMBadIntHandler

              POP       AX                  ; recover the interrupt ID
              NOT       AX                  ; flip it to make it readable
              MOV       DX,LED_ADDR         ; set up for the display

              XOR       CX,CX               ; set up initial count

@@Stall:
              OUT       DX,AX               ; display new data
              XOR       AX,08000h           ; flip the high decimal point
              MOV       BL,5
@@Punt:       LOOP      @@Punt              ; delay for a decent interval
              DEC       BL
              JNZ       @@Punt
              JMP       @@Stall

              ENDP      PMBadIntHandler

              END
